import unittest
from datetime import datetime

import pandas as pd
from pandas.testing import assert_frame_equal
from parameterized import parameterized

from dgraphpandas.strategies.vertical import vertical_transform


class Vertical(unittest.TestCase):

    @parameterized.expand([
        (None, {'config', 'fake'}, 'config'),
        (pd.DataFrame(), None, 'config'),
        (pd.DataFrame(), {'config', 'fake'}, None),
        (pd.DataFrame(), {'config', 'fake'}, None),
        (pd.DataFrame(), {}, 'config'),
        (pd.DataFrame(), None, '')
    ])
    def test_vertical_transform_null_parameter(self, frame, config, config_file_key):
        '''
        Ensures when parameters are null, an error is raised.
        '''
        with self.assertRaises(ValueError):
            vertical_transform(frame, config, config_file_key)

    def test_vertical_transform_config_file_key_not_in_config(self):
        '''
        Ensures when the passed config key is not within the config
        then an error is raised.
        '''
        frame = pd.DataFrame()
        config_file_key = 'not_my_file'
        config = {
            'files': {'my_file': {}}
        }
        with self.assertRaises(KeyError):
            vertical_transform(frame, config, config_file_key)

    def test_vertical_transform_subject_fields_not_defined(self):
        '''
        Ensures when subject_fields has not been defined,
        then an error is raised.
        '''
        frame = pd.DataFrame()
        config_file_key = 'customer'
        config = {
            'files': {
                config_file_key: {
                    'type_overrides': {}
                }
            }
        }

        with self.assertRaises(ValueError):
            vertical_transform(frame, config, config_file_key)

    def test_vertical_predicate_column_not_defined(self):
        '''
        Ensures when the predicate column is not defined
        then we throw an error.
        '''
        frame = pd.DataFrame(data={
            'customer_id': [1, 2, 3],
            'not_a_predicate': ['age', 'weight', 'orders'],
            'object': [23, 90, 10]
        })

        config_file_key = 'customer'
        config = {
            'files': {
                'customer': {
                    'subject_fields': ['customer_id'],
                }
            }
        }

        with self.assertRaises(KeyError):
            vertical_transform(frame, config, config_file_key)

    def test_vertical_object_column_not_defined(self):
        '''
        Ensures when the object column is not defined
        then we throw an error.
        '''
        frame = pd.DataFrame(data={
            'customer_id': [1, 2, 3],
            'predicate': ['age', 'weight', 'orders'],
            'not_a_object': [23, 90, 10]
        })

        config_file_key = 'customer'
        config = {
            'files': {
                'customer': {
                    'subject_fields': ['customer_id'],
                }
            }
        }

        with self.assertRaises(KeyError):
            vertical_transform(frame, config, config_file_key)

    def test_vertical_transform_intrinsic_with_default_values(self):
        '''
        Ensures when a DataFrame is passed with the default options
        then an intrinsic dataframe is created
        and an empty edges frame is created.
        '''
        frame = pd.DataFrame(data={
            'customer_id': [1, 2, 3],
            'predicate': ['age', 'weight', 'orders'],
            'object': [23, 90, 10]
        })

        config_file_key = 'customer'
        config = {
            'files': {
                'customer': {
                    'subject_fields': ['customer_id'],
                    'dgraph_type': "customer"
                }
            }
        }

        expected_frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_2', 'customer_3', 'customer_1', 'customer_2', 'customer_3'],
            'predicate': ['age', 'weight', 'orders', 'dgraph.type', 'dgraph.type', 'dgraph.type'],
            'object': [23, 90, 10, 'customer', 'customer', 'customer'],
            'type': ['<xs:string>']*6
        })

        intrinsic, edges = vertical_transform(frame, config, config_file_key)

        intrinsic = intrinsic.reset_index(drop=True)
        expected_frame = expected_frame.reset_index(drop=True)

        assert_frame_equal(expected_frame, intrinsic)
        self.assertTrue(edges.empty)

    def test_vertical_transform_with_type_overrides(self):
        '''
        Ensures when type overrides have been provided, they are
        applied
        '''
        frame = pd.DataFrame(data={
            'customer_id': [1, 2, 3],
            'predicate': ['age', 'weight', 'orders'],
            'object': [23, 90, 10]
        })

        config_file_key = 'customer'
        config = {
            'files': {
                'customer': {
                    'subject_fields': ['customer_id'],
                    'dgraph_type': "customer",
                    'type_overrides': {
                        'age': 'int32',
                        'weight': 'float32',
                        'orders': 'int32',
                    }
                }
            }
        }

        expected_frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_2', 'customer_3', 'customer_1', 'customer_2', 'customer_3'],
            'predicate': ['age', 'weight', 'orders', 'dgraph.type', 'dgraph.type', 'dgraph.type'],
            'object': [23, 90, 10, 'customer', 'customer', 'customer'],
            'type': ['<xs:int>', '<xs:float>', '<xs:int>'] + ['<xs:string>']*3
        })

        intrinsic, edges = vertical_transform(frame, config, config_file_key)

        intrinsic = intrinsic.reset_index(drop=True)
        expected_frame = expected_frame.reset_index(drop=True)

        assert_frame_equal(expected_frame, intrinsic)
        self.assertTrue(edges.empty)

    @parameterized.expand([
        ###
        (
            'with_default_values',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                        'dgraph_type': "customer",
                        'type_overrides': {
                            'age': 'int32',
                            'weight': 'float32',
                            'orders': 'int32',
                        }
                    }
                }
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, 90, 10]
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2', 'customer_3', 'customer_1', 'customer_2', 'customer_3'],
                'predicate': ['age', 'weight', 'orders', 'dgraph.type', 'dgraph.type', 'dgraph.type'],
                'object': [23, 90, 10, 'customer', 'customer', 'customer'],
                'type': ['<xs:int>', '<xs:float>', '<xs:int>'] + ['<xs:string>']*3
            }),
            pd.DataFrame(columns=['subject', 'predicate', 'object', 'type']),
            {}
        ),
        ###
        (
            'no_dgraph_type_defaults_to_config_key',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                        'type_overrides': {
                            'age': 'int32',
                            'weight': 'float32',
                            'orders': 'int32',
                        }
                    }
                }
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, 90, 10]
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2', 'customer_3', 'customer_1', 'customer_2', 'customer_3'],
                'predicate': ['age', 'weight', 'orders', 'dgraph.type', 'dgraph.type', 'dgraph.type'],
                'object': [23, 90, 10, 'customer', 'customer', 'customer'],
                'type': ['<xs:int>', '<xs:float>', '<xs:int>'] + ['<xs:string>']*3
            }),
            pd.DataFrame(columns=['subject', 'predicate', 'object', 'type']),
            {}
        ),
        ###
        (
            'rename_fields',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                        'pre_rename': {
                            'orders': 'no_of_orders',
                            'age': 'birth'
                        }
                    }
                }
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, 90, 10]
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2', 'customer_3', 'customer_1', 'customer_2', 'customer_3'],
                'predicate': ['birth', 'weight', 'no_of_orders', 'dgraph.type', 'dgraph.type', 'dgraph.type'],
                'object': [23, 90, 10, 'customer', 'customer', 'customer'],
                'type': ['<xs:string>']*6
            }),
            pd.DataFrame(columns=['subject', 'predicate', 'object', 'type']),
            {}
        ),
        ###
        (
            'rename_fields_kwargs',
            'customer',
            {
                'files': {
                    'customer': {'subject_fields': ['customer_id']}
                }
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, 90, 10]
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2', 'customer_3', 'customer_1', 'customer_2', 'customer_3'],
                'predicate': ['birth', 'weight', 'no_of_orders', 'dgraph.type', 'dgraph.type', 'dgraph.type'],
                'object': [23, 90, 10, 'customer', 'customer', 'customer'],
                'type': ['<xs:string>']*6
            }),
            pd.DataFrame(columns=['subject', 'predicate', 'object', 'type']),
            {
                'pre_rename': {
                    'orders': 'no_of_orders',
                    'age': 'birth'
                }
            }
        ),
        ###
        (
            'ignore_fields',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                        'ignore_fields': ['age']
                    }
                }
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, 90, 10]
            }),
            pd.DataFrame(data={
                'subject': ['customer_2', 'customer_3', 'customer_2', 'customer_3'],
                'predicate': ['weight', 'orders', 'dgraph.type', 'dgraph.type'],
                'object': [90, 10, 'customer', 'customer'],
                'type': ['<xs:string>']*4
            }),
            pd.DataFrame(columns=['subject', 'predicate', 'object', 'type']),
            {}
        ),
        ###
        (
            'ignore_fields_kwargs',
            'customer',
            {
                'files': {
                    'customer': {'subject_fields': ['customer_id']}
                }
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, 90, 10]
            }),
            pd.DataFrame(data={
                'subject': ['customer_2', 'customer_3', 'customer_2', 'customer_3'],
                'predicate': ['weight', 'orders', 'dgraph.type', 'dgraph.type'],
                'object': [90, 10, 'customer', 'customer'],
                'type': ['<xs:string>']*4
            }),
            pd.DataFrame(columns=['subject', 'predicate', 'object', 'type']),
            {'ignore_fields': ['age']}
        ),
        ###
        (
            'with_csv_edges',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                        'csv_edges': ['orders']
                    },
                },
                'add_dgraph_type_records': False
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, 90, '1,2,3']
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2', 'customer_3', 'customer_3', 'customer_3'],
                'predicate': ['age', 'weight', 'orders', 'orders', 'orders'],
                'object': [23, 90, '1', '2', '3'],
                'type': ['<xs:string>']*5
            }),
            pd.DataFrame(columns=['subject', 'predicate', 'object', 'type']),
            {}
        ),
        ###
        (
            'with_csv_edges_kwargs',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                    },
                },
                'add_dgraph_type_records': False
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, 90, '1,2,3']
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2', 'customer_3', 'customer_3', 'customer_3'],
                'predicate': ['age', 'weight', 'orders', 'orders', 'orders'],
                'object': [23, 90, '1', '2', '3'],
                'type': ['<xs:string>']*5
            }),
            pd.DataFrame(columns=['subject', 'predicate', 'object', 'type']),
            {'csv_edges': ['orders']}
        ),
        ###
        (
            'with_composite_key',
            'order',
            {
                'files': {
                    'order': {
                        'subject_fields': ['customer_id', 'order_id']
                    }
                },
                'add_dgraph_type_records': False
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3],
                'order_id': [908, 456, 287],
                'predicate': ['total_value', 'quantity', 'item'],
                'object': [2031, 5, 'ipads']
            }),
            pd.DataFrame(data={
                'subject': ['order_1_908', 'order_2_456', 'order_3_287'],
                'predicate': ['total_value', 'quantity', 'item'],
                'object': [2031, 5, 'ipads'],
                'type': ['<xs:string>']*3
            }),
            pd.DataFrame(columns=['subject', 'predicate', 'object', 'type']),
            {}
        ),
        ###
        (
            'with_edges',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                        'edge_fields': ['location_id']
                    },
                },
                'add_dgraph_type_records': False
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3, 1, 2],
                'predicate': ['age', 'weight', 'orders', 'location_id', 'location_id'],
                'object': [23, 90, '1', 'loc45', 'loc64']
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2', 'customer_3'],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, 90, '1'],
                'type': ['<xs:string>']*3
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2'],
                'predicate': ['location', 'location'],
                'object': ['location_loc45', 'location_loc64'],
                'type': [None]*2
            }),
            {}
        ),
        ###
        (
            'strip_id_from_edge_names',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                        'edge_fields': ['location_id']
                    },
                },
                'strip_id_from_edge_names': False,
                'add_dgraph_type_records': False,
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3, 1, 2],
                'predicate': ['age', 'weight', 'orders', 'location_id', 'location_id'],
                'object': [23, 90, '1', 'loc45', 'loc64']
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2', 'customer_3'],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, 90, '1'],
                'type': ['<xs:string>']*3
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2'],
                'predicate': ['location_id', 'location_id'],
                'object': ['location_id_loc45', 'location_id_loc64'],
                'type': [None]*2
            }),
            {}
        ),
        ###
        (
            'strip_id_from_edge_names_kwargs',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                        'edge_fields': ['location_id']
                    },
                },
                'strip_id_from_edge_names': False
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3, 1, 2],
                'predicate': ['age', 'weight', 'orders', 'location_id', 'location_id'],
                'object': [23, 90, '1', 'loc45', 'loc64']
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2', 'customer_3'],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, 90, '1'],
                'type': ['<xs:string>']*3
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2'],
                'predicate': ['location_id', 'location_id'],
                'object': ['location_id_loc45', 'location_id_loc64'],
                'type': [None]*2
            }),
            {'add_dgraph_type_records': False}
        ),
        ###
        (
            'key_seperator override',
            'order',
            {
                'files': {
                    'order': {
                        'subject_fields': ['customer_id', 'order_id']
                    }
                },
                'add_dgraph_type_records': False,
                'key_separator': "@"
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3],
                'order_id': [908, 456, 287],
                'predicate': ['total_value', 'quantity', 'item'],
                'object': [2031, 5, 'ipads']
            }),
            pd.DataFrame(data={
                'subject': ['order@1@908', 'order@2@456', 'order@3@287'],
                'predicate': ['total_value', 'quantity', 'item'],
                'object': [2031, 5, 'ipads'],
                'type': ['<xs:string>']*3
            }),
            pd.DataFrame(columns=['subject', 'predicate', 'object', 'type']),
            {}
        ),
        ###
        (
            'with datetime fields',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                        'edge_fields': ['location_id'],
                        'type_overrides': {
                            'dob': 'datetime64'
                        }
                    },
                },
                'add_dgraph_type_records': False,
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3, 1, 2],
                'predicate': ['dob', 'weight', 'orders', 'location_id', 'location_id'],
                'object': [datetime(2021, 4, 5), 90, '1', 'loc45', 'loc64']
            }),
            pd.DataFrame(data={
                'subject': ['customer_2', 'customer_3', 'customer_1'],
                'predicate': ['weight', 'orders', 'dob'],
                'object': [90, '1', '2021-04-05T00:00:00'],
                'type': ['<xs:string>']*2 + ['<xs:dateTime>']
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2'],
                'predicate': ['location', 'location'],
                'object': ['location_loc45', 'location_loc64'],
                'type': [None]*2
            }),
            {}
        ),
        ###
        (
            'illegal_characters',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                        'edge_fields': ['location_id']
                    },
                },
                'illegal_characters': ['%'],
                'illegal_characters_intrinsic_object': ['\"'],
                'add_dgraph_type_records': False,
            },
            pd.DataFrame(data={
                'customer_id': ['%1%', 2, 3, 1, 2],
                'predicate': ['age', 'weight', 'orders', 'location_id', 'location_id'],
                'object': [23, '"90"', '1', 'loc45', 'loc64']
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2', 'customer_3'],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, '90', '1'],
                'type': ['<xs:string>']*3
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2'],
                'predicate': ['location', 'location'],
                'object': ['location_loc45', 'location_loc64'],
                'type': [None]*2
            }),
            {}
        ),
        ###
        (
            'null_objects',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                        'edge_fields': ['location_id']
                    },
                },
                'add_dgraph_type_records': False,
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3, 1, 2],
                'predicate': ['age', 'weight', 'orders', 'location_id', 'location_id'],
                'object': [23, None, '1', 'loc45', None]
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_3'],
                'predicate': ['age', 'orders'],
                'object': [23, '1'],
                'type': ['<xs:string>']*2
            }),
            pd.DataFrame(data={
                'subject': ['customer_1'],
                'predicate': ['location'],
                'object': ['location_loc45'],
                'type': [None]
            }),
            {}
        ),
        ###
        (
            'override_edge_name',
            'customer',
            {
                'files': {
                    'customer': {
                        'subject_fields': ['customer_id'],
                        'edge_fields': ['location_id'],
                        'override_edge_name': {
                            'location': {
                                'predicate': 'habitat',
                                'target_node_type': 'hab'
                            }
                        }
                    },
                },
                'add_dgraph_type_records': False
            },
            pd.DataFrame(data={
                'customer_id': [1, 2, 3, 1, 2],
                'predicate': ['age', 'weight', 'orders', 'location_id', 'location_id'],
                'object': [23, 90, '1', 'loc45', 'loc64']
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2', 'customer_3'],
                'predicate': ['age', 'weight', 'orders'],
                'object': [23, 90, '1'],
                'type': ['<xs:string>']*3
            }),
            pd.DataFrame(data={
                'subject': ['customer_1', 'customer_2'],
                'predicate': ['habitat', 'habitat'],
                'object': ['hab_loc45', 'hab_loc64'],
                'type': [None]*2
            }),
            {'add_dgraph_type_records': False}
        ),

    ])
    def test_vertical_transform(
            self,
            name,
            config_file_key,
            config,
            input_frame,
            expected_intrinsic,
            expected_edges,
            kwargs):
        '''
        Ensures that Vertical Transforms are working.
        Each of the parameter sets being passed here represents:
            For a given input frame and configuration,
            does the output match the given intrinsic and edges

        Note that the core flow of running the logic is the same but
        all the different combination of inputs is coming from the parameters.

        add_dgraph_type_records is added in some of these tests only to reduce
        the verbosity of our DataFrame declarations.
        '''
        intrinsic, edges = vertical_transform(input_frame, config, config_file_key, **kwargs)

        intrinsic = intrinsic.reset_index(drop=True)
        expected_intrinsic = expected_intrinsic.reset_index(drop=True)

        edges = edges.reset_index(drop=True)
        expected_edges = expected_edges.reset_index(drop=True)

        try:
            assert_frame_equal(expected_intrinsic, intrinsic)
            assert_frame_equal(expected_edges, edges)
        except AssertionError:
            print('Failed', name)
            print('Input:')
            print(input_frame)
            print('Expected Intrinsic:')
            print(expected_intrinsic)
            print('Intrinsic:')
            print(intrinsic)
            print('#####')
            print('Expected Edges:')
            print(expected_edges)
            print('Edges:')
            print(edges)
            raise



import re
import logging
from typing import Callable, Dict, Any, List, Pattern, Tuple, Union

import pandas as pd

from dgraphpandas.types import default_rdf_type
from dgraphpandas.types import find_rdf_types


logger = logging.getLogger(__name__)


def _expand_csv_edges(frame: pd.DataFrame, csv_edges: List[str], seperator=',') -> pd.DataFrame:
    '''
    Sometimes fields will be delivered in CSV format.
    For example a Movie record might have a cast field with the value: actor1, actor2, actor3
    In this kind of situation, we want the actors to be broken up into 3 different nodes

    When a field is declared within csv_edges, then we break up each of the values in the csv into
    it's own record.
    '''

    if frame is None:
        raise ValueError('frame')

    if csv_edges:
        logger.info(f'Detected csv_edges {csv_edges}. Breaking up those columns')
        csv_edge_frame = frame[frame['predicate'].isin(csv_edges)]
        csv_edge_frame['object'] = csv_edge_frame['object'].str.split(seperator)
        csv_edge_frame = csv_edge_frame.explode(column='object')
        csv_edge_frame.dropna(subset=['object'], inplace=True)
        csv_edge_frame['object'] = csv_edge_frame['object'].str.strip()

        frame = frame[~(frame['predicate'].isin(csv_edges))]
        frame = pd.concat([frame, csv_edge_frame])

    return frame


def _join_key_fields(frame: pd.DataFrame, key: List[str], key_seperator: str, type: str) -> pd.DataFrame:
    '''
    If we have a composite key, then join all into a single subject column
    We do this length check as a performance optimisation as apply can take some time
    and there is no need if there is only 1 key.
    '''
    if frame is None:
        raise ValueError('frame')
    elif key is None:
        raise ValueError('key')
    elif key_seperator is None:
        raise ValueError('key_seperator')
    elif type is None:
        raise ValueError('type')

    logger.debug(f'Joining Key fields {key} to subject')
    if len(key) > 1:
        frame['subject'] = frame[key].apply(lambda row: key_seperator.join(row.values.astype(str)), axis=1)
    else:
        frame[key] = frame[key].astype(str)
        frame['subject'] = frame[key]

    frame['subject'] = type + key_seperator + frame['subject']

    logger.debug('Dropping keys in favour of subject')
    frame = frame.drop(labels=key, axis=1)
    return frame


def _add_dgraph_type_records(frame: pd.DataFrame, add_dgraph_type_records: bool, type: str) -> pd.DataFrame:
    '''
    Dgraph has a special field called 'dgraph.type', this can be used to query via the type()
    function. If add_dgraph_type_records is enabled, then we add dgraph.type fields
    to the current frame.
    '''
    if add_dgraph_type_records:
        logger.debug('Adding dgraph.type fields')
        add_type_frame = frame.copy()
        add_type_frame['object'] = type
        add_type_frame['predicate'] = 'dgraph.type'
        frame = pd.concat([frame, add_type_frame])

    return frame


def _break_up_intrinsic_and_edges(frame: pd.DataFrame, edges: List[str], strip_id_from_edge_names: bool = True) -> Tuple[pd.DataFrame, pd.DataFrame]:
    '''
    If there are edges defined, then break up into intrinsic and edge frames
    Otherwise still break up but return an empty edges frame

    For strip_id_from_edge_names, Its common for a data set to have a reference to another 'table' using _id
    convention. For example if you had a Student & School then the student might
    have a 'school_id' field which points to the school. In a Graph Database it might make
    more sense to have (Student) - school -> (School) rather then having an _id in the predicate.
    '''
    if frame is None:
        raise ValueError('frame')

    if edges:
        logger.debug(f'Splitting into Intrinsic and edges based on edges {edges}')
        intrinsic = frame.loc[~frame['predicate'].isin(edges)]
        edges = frame.loc[frame['predicate'].isin(edges)]

        if strip_id_from_edge_names:
            edges['predicate'] = edges['predicate'].str.replace('_id', '')

    else:
        logger.debug('No Edges defined, Skipping.')
        intrinsic = frame
        edges = pd.DataFrame(columns=['subject', 'predicate', 'object', 'type'])

    return intrinsic, edges


def _apply_rdf_types(frame: pd.DataFrame, types: Dict[str, str]):
    '''
    For each of the vertical rows, map each predicate to a pre-defined rdf type.
    Edges are always uid type, so don't bother filling this in.

    For any types which could not be applied, then default to default_rdf_type
    '''

    if frame is None:
        raise ValueError('frame')
    if types is None:
        raise ValueError('types')

    logger.debug('Applying RDF Types')
    rdf_types = find_rdf_types(types)
    frame['type'] = frame['predicate'].map(rdf_types)
    frame['type'].fillna(default_rdf_type, inplace=True)

    return frame


def _format_date_fields(frame: pd.DataFrame) -> pd.DataFrame:
    '''
    Ensure that DateTime fields are formatted in ISO format
    And any fields are which NaT are filtered out.
    '''
    if frame is None:
        raise ValueError('frame')

    logger.debug('Ensuring Date Time fields are in ISO format')
    intrinsic_with_datetime = frame.loc[frame['type'] == '<xs:dateTime>']
    frame = frame.loc[frame['type'] != '<xs:dateTime>']

    try:
        intrinsic_with_datetime['object'] = intrinsic_with_datetime['object'].apply(lambda x: x.isoformat())
    except AttributeError as e:
        logger.error(
            'It looks like a value being declared as a datetime is not actually a datetime',
            e,
            extra={'frame': intrinsic_with_datetime})
        raise

    intrinsic_with_datetime = intrinsic_with_datetime.loc[intrinsic_with_datetime['object'] != 'NaT']
    frame = pd.concat([frame, intrinsic_with_datetime])
    return frame


def _compile_illegal_characters_regex(characters: List[str]) -> Union[Pattern, None]:
    if not characters:
        return None

    return re.compile('|'.join(characters))


def _remove_illegal_rdf_characters(frame: pd.DataFrame, illegal_characters: Union[List[str], Pattern], field: str):
    '''
    Some characters are illegal in an RDF export and DGraph will not accept them.
    So make sure we filter out any of those characters from our Frame.

    There is a differentiation between illegal characters in rdf in general and the values
    which can be inside the subject fields for intrinsic values since these are quoted and are
    actual data items which may be visible to clients.
    '''

    if frame is None:
        raise ValueError('frame')
    if not field:
        raise ValueError('field')

    logger.debug('Compiling Illegal Characters %s', illegal_characters)
    if isinstance(illegal_characters, list):
        logger.debug('Resolving illegal_characters')
        illegal_characters: Pattern = _compile_illegal_characters_regex(illegal_characters)

    if illegal_characters:
        frame[field] = frame[field].replace(illegal_characters, '')

    return frame


def _remove_na_objects(frame: pd.DataFrame, drop_na: bool = True):
    '''
    If the object is NA/Null then the predicate does not exist from this node
    to the target node, so drop those records.
    '''
    if frame is None:
        raise ValueError('frame')

    if drop_na:
        logger.debug('Dropping records where NA on object')
        frame.dropna(subset=['object'], inplace=True)

    return frame


def _override_edge_name(edges: pd.DataFrame, override_edge_name: Dict[str, Any], key_seperator: str):
    '''
    For Edges, we want to make sure the objects have the predicate prepended
    so that they map to the corresponding xid for the actual node.

    For example if we had a Student / School relationship and we are
    currently operating on the student then we might have
    <student_45> <school> <1> but we really want <student_45> <school> <school_1>
    because school_1 would be found whenever we are loading our school data.

    If override_edge_name has been specified then do some extra work
    to make the predicate and target object type different.
    '''
    if edges is None:
        raise ValueError('edges')
    if not key_seperator:
        raise ValueError('key_seperator')

    def _override_edge_name_apply(row: pd.Series, override_edge_name: Dict[str, Any], key_seperator: str):
        if row['predicate'] not in override_edge_name:
            row['object'] = row['predicate'] + key_seperator + str(row['object'])
        else:
            current_override = override_edge_name[row['predicate']]
            row['predicate'] = current_override['predicate'] if 'predicate' in current_override else row['predicate']
            row['object'] = current_override['target_node_type'] + key_seperator + str(row['object'])

        return row

    if override_edge_name is not None and any(override_edge_name):
        edges = edges.apply(_override_edge_name_apply, axis='columns', args=(override_edge_name, key_seperator))
    else:
        edges['object'] = edges['predicate'].astype(str) + key_seperator + edges['object'].astype(str)

    return edges


def _ignore_fields(frame: pd.DataFrame, ignore_fields: List[str]) -> pd.DataFrame:
    '''
    For any predicates within the ignore fields list, remove them
    from our frame.
    '''
    if frame is None:
        raise ValueError('frame')

    if ignore_fields:
        frame = frame[~frame['predicate'].isin(ignore_fields)]

    return frame


def _resolve_potential_callables(frame: pd.DataFrame, potential_callables: Dict[str, Union[List[str], str, Callable]]) -> Dict[str, Union[List[str], str, Callable]]:
    '''
    Some user defined parameters are Callable so that they can be derived from the frame dynamically.
    For example, if you wanted to apply a convention that all your edges always end with _id
    then for the edge_fields parameter, you could provide a lambda which resolves into a list given the frame
    '''
    if frame is None:
        raise ValueError('frame')

    for key, potential_callable in potential_callables.items():
        if callable(potential_callable):
            logger.debug('resolving %s', key)
            potential_callables[key] = potential_callable(frame)
            logger.debug('Resolved %s to %s', key, potential_callables[key])

    return potential_callables


def _rename_fields(frame: pd.DataFrame, pre_rename: Dict[str, str]) -> pd.DataFrame:
    '''
    For any renames defined in the incoming pre_rename, then update the predicate names
    according to that mapping. If the predicate does not exist in the mapping
    then leave untouched.
    '''
    if frame is None:
        raise ValueError('frame')

    if pre_rename:
        frame['predicate'] = frame['predicate'].apply(lambda x: pre_rename.get(x, x))

    return frame



import unittest
from parameterized import parameterized
from dgraphpandas.config import get_from_config


class ConfigTests(unittest.TestCase):

    @parameterized.expand([
        (None, {'hello': 'world'}, 'default', {'firstkey': 'secondkey'}),
        ('key', None, 'default', {'firstkey': 'secondkey'})
    ])
    def test_get_from_config_nullkey_exception(
            self,
            key,
            config,
            default,
            kwargs):
        '''
        Ensures when the passed key is none, then an exception is raised.
        '''
        with self.assertRaises(ValueError):
            get_from_config(key, config, default, **(kwargs))

    def test_get_from_config_override_exists(self):
        '''
        Ensures when an override exists in kwargs, then it is used
        even if a value exists in config
        '''
        key = 'subject_fields'
        config = {'subject_fields': 'id'}
        value = get_from_config(key, config, subject_fields='override')

        self.assertEqual(value, 'override')

    def test_get_from_config_no_override_exists(self):
        '''
        Ensures when no override exists in kwargs, then the config value is used
        '''
        key = 'subject_fields'
        config = {'subject_fields': 'id'}
        value = get_from_config(key, config, something_else='override')

        self.assertEqual(value, 'id')

    def test_get_from_config_no_override_and_no_config_default(self):
        '''
        Ensures when no override exists in kwargs, then the config value is used
        '''
        key = 'subject_fields'
        config = {'some_other_field': 'value'}
        value = get_from_config(key, config, something_else='override', default='default_value')

        self.assertEqual(value, 'default_value')



import logging
import argparse
import os
import gzip
import json
from typing import Any, Dict, List
from pprint import pformat

import pandas as pd

from dgraphpandas.strategies.horizontal import horizontal_transform
from dgraphpandas.strategies.vertical import vertical_transform
from dgraphpandas.writers.upserts import generate_upserts

logger = logging.getLogger(__name__)

try:
    import coloredlogs
    coloredlogs.install(level='DEBUG')
except ImportError as e:
    logger.warning(e)

pd.set_option('mode.chained_assignment', None)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--file')
    parser.add_argument('-c', '--config')
    parser.add_argument('-ck', '--config_file_key')
    parser.add_argument('-o', '--output_dir', default='.')
    parser.add_argument('--console', action='store_true', default=False)
    parser.add_argument('--pre_csv', action='store_true', default=False)
    parser.add_argument('--skip_upsert_generation', action='store_true', default=False)
    parser.add_argument('--encoding', default=os.environ.get('DGRAPH_PANDAS_ENCODING', 'utf-8'))

    args = parser.parse_args()

    with open(args.config, 'r') as f:
        global_config: Dict[str, Any] = json.load(f)
        logger.debug('Global Config \n %s', pformat(global_config))

    all_intrinsic: List[pd.DataFrame] = []
    all_edges: List[pd.DataFrame] = []

    if global_config['transform'] == 'horizontal':
        intrinsic, edges = horizontal_transform(args.file, global_config, args.config_file_key)
        all_intrinsic.append(intrinsic)
        all_edges.append(edges)

    elif global_config['transform'] == 'vertical':
        intrinsic, edges = vertical_transform(args.file, global_config, args.config_file_key)
        all_intrinsic.append(intrinsic)
        all_edges.append(edges)

    else:
        raise NotImplementedError(global_config['transform'])

    logger.debug(f'Concatting {len(all_intrinsic)} intrinsic frames and {len(all_edges)} frames')
    intrinsic = pd.concat(all_intrinsic)
    edges = pd.concat(all_edges)

    if args.console:
        print('Intrinsic:')
        print(intrinsic)
        print('Edges:')
        print(edges)

    if args.pre_csv:
        os.makedirs(args.output_dir, exist_ok=True)
        source_file_name = os.path.basename(args.file).split('.')[0]

        intrinsic_path = os.path.join(args.output_dir, source_file_name + '_intrinsic.csv')
        edges_path = os.path.join(args.output_dir, source_file_name + '_edges.csv')

        logger.info(f'Writing to {intrinsic_path}')
        intrinsic.to_csv(intrinsic_path, index=False, encoding=args.encoding)

        logger.info(f'Writing to {edges_path}')
        edges.to_csv(edges_path, index=False, encoding=args.encoding)

    if not args.skip_upsert_generation:
        logger.info('Generating Upsert Queries')
        intrinsic_upserts, edges_upserts = generate_upserts(intrinsic, edges)

        os.makedirs(args.output_dir, exist_ok=True)
        intrinsic_path = os.path.join(args.output_dir, os.path.basename(args.file).split('.')[0] + '_intrinsic.gz')
        edges_path = os.path.join(args.output_dir, os.path.basename(args.file).split('.')[0] + '_edges.gz')

        logger.info(f'Writing to {len(intrinsic_upserts)} upserts to {intrinsic_path}')
        with gzip.open(intrinsic_path, mode='wb', compresslevel=9) as zip:
            s = '\n'.join(intrinsic_upserts)
            s = s.encode(encoding=args.encoding)
            zip.write(s)

        logger.info(f'Writing to {len(edges_upserts)} upserts to {edges_path}')
        with gzip.open(edges_path, mode='wb', compresslevel=9) as zip:
            s = '\n'.join(edges_upserts)
            s = s.encode(encoding=args.encoding)
            zip.write(s)
    else:
        logger.warning('skip_upsert_generation was set, skipping')


if __name__ == '__main__':
    main()



import unittest

from dgraphpandas.types import find_rdf_types, default_rdf_type


class TypesTests(unittest.TestCase):

    def test_find_rdf_types_null_types_exception(self):
        '''
        Ensures when the types passed are null,
        then a value error is raised.
        '''
        with self.assertRaises(ValueError):
            find_rdf_types(None)

    def test_find_rdf_types_some_types_resolvable(self):
        '''
        Ensures when some of the types are resolvable
        to rdf those are resolved but the unresolvable
        are given the default rdf type.
        '''
        types = {
            'column1': 'int32',
            'column2': 'float32',
            'column3': 'something_unknown'
        }

        expected = {
            'column1': '<xs:int>',
            'column2': '<xs:float>',
            'column3': default_rdf_type
        }

        resolved_types = find_rdf_types(types)
        self.assertEqual(expected, resolved_types)

    def test_find_rdf_types_all_resolvable(self):
        '''
        Ensures when all the types are resolvable,
        they are all resolved.
        '''
        types = {
                'column1': 'object',
                'column2': 'O',
                'column3': 'int32',
                'column4': 'int64',
                'column5': 'int32',
                'column6': 'int64',
                'column7': 'float32',
                'column8': 'float64',
                'column9': 'float32',
                'column10': 'float64',
                'column11': 'datetime64',
                'column12': '<M8[ns]',
                'column13': 'bool'
        }

        expected = {
                'column1': '<xs:string>',
                'column2': '<xs:string>',
                'column3': '<xs:int>',
                'column4': '<xs:int>',
                'column5': '<xs:int>',
                'column6': '<xs:int>',
                'column7': '<xs:float>',
                'column8': '<xs:float>',
                'column9': '<xs:float>',
                'column10': '<xs:float>',
                'column11': '<xs:dateTime>',
                'column12': '<xs:dateTime>',
                'column13': '<xs:bool>'
        }

        resolved_types = find_rdf_types(types)
        self.assertEqual(expected, resolved_types)

    def test_find_rdf_types_none_resolvable(self):
        '''
        Ensures when none of the columns can be resolved,
        they are all resolved to default rdf type
        '''
        types = {
                'column1': 'unknown',
                'column2': 'unknown2',
                'column3': 'unknown3',
                'column4': 'unknown4'
        }

        resolved_types = find_rdf_types(types)
        self.assertEqual(list(resolved_types.values()), [default_rdf_type]*4)



import logging
from typing import Callable, Dict, Any, List, Union

import pandas as pd

from dgraphpandas.config import get_from_config
from dgraphpandas.strategies.vertical_helpers import (_expand_csv_edges, _join_key_fields, _add_dgraph_type_records,
                                                      _break_up_intrinsic_and_edges, _apply_rdf_types, _format_date_fields,
                                                      _remove_illegal_rdf_characters, _remove_na_objects, _override_edge_name,
                                                      _ignore_fields, _resolve_potential_callables, _rename_fields)


logger = logging.getLogger(__name__)


def vertical_transform(
        frame: Union[str, pd.DataFrame],
        config: Dict[str, Any],
        config_file_key: str,
        **kwargs):
    '''
    Vertically Transform a Pandas Dataframe into Intrinsic and Edge DataFrames (close to RDF format)
    '''
    if frame is None:
        raise ValueError('frame')
    if not config:
        raise ValueError('config')
    if not config_file_key:
        raise ValueError('config_file_key')

    try:
        file_config: Dict[str, Any] = config['files'][config_file_key]
    except KeyError:
        logger.exception(f'Ensure that {config_file_key} is within the files object in config')
        raise

    if isinstance(frame, str):
        logger.debug(f'Reading file {frame}')
        frame = pd.read_csv(frame)

    subject_fields: Union[List[str], Callable[..., List[str]]] = get_from_config('subject_fields', file_config, **(kwargs))
    edge_fields: Union[List[str], Callable[..., List[str]]] = get_from_config('edge_fields', file_config, [], **(kwargs))
    dgraph_type: str = get_from_config('dgraph_type', file_config, config_file_key, **(kwargs))

    key_seperator: str = get_from_config('key_separator', config, '_', **(kwargs))
    add_dgraph_type_records: bool = get_from_config('add_dgraph_type_records', config, True, **(kwargs))
    strip_id_from_edge_names: bool = get_from_config('strip_id_from_edge_names', config, True, **(kwargs))
    drop_na_intrinsic_objects: bool = get_from_config('drop_na_intrinsic_objects', config, True, **(kwargs))
    drop_na_edge_objects: bool = get_from_config('drop_na_edge_objects', config, True, **(kwargs))
    illegal_characters: List[str] = get_from_config('illegal_characters', config, ['%', '\\.', '\\s', '\"', '\\n', '\\r\\n'], **(kwargs))
    illegal_characters_intrinsic_object: List[str] = get_from_config('illegal_characters_intrinsic_object', config, ['\"', '\\n', '\\r\\n'], **(kwargs))
    csv_edges: str = get_from_config('csv_edges', file_config, [], **(kwargs))
    ignore_fields: List[str] = get_from_config('ignore_fields', file_config, [], **(kwargs))
    override_edge_name: Dict[str, Any] = get_from_config('override_edge_name', file_config, {}, **(kwargs))
    pre_rename: Dict[str, str] = get_from_config('pre_rename', file_config, {}, **(kwargs))
    type_overrides: Dict[str, str] = get_from_config('type_overrides', file_config, {}, **(kwargs))

    potential_callables = _resolve_potential_callables(frame, {
        'subject_fields': subject_fields,
        'edge_fields': edge_fields,
        'dgraph_type': dgraph_type
    })

    key = potential_callables["subject_fields"]
    edges = potential_callables['edge_fields']
    type = potential_callables['dgraph_type']

    if not key:
        raise ValueError('subject_fields must be defined')
    if 'predicate' not in frame.columns:
        raise KeyError('predicate column must be defined on vertical frame')
    if 'object' not in frame.columns:
        raise KeyError('object column must be defined on vertical frame')

    frame = _rename_fields(frame, pre_rename)
    frame = _ignore_fields(frame, ignore_fields)
    frame = _expand_csv_edges(frame, csv_edges)
    frame = _join_key_fields(frame, key, key_seperator, type)
    frame = _add_dgraph_type_records(frame, add_dgraph_type_records, type)

    intrinsic, edges = _break_up_intrinsic_and_edges(frame, edges, strip_id_from_edge_names)
    intrinsic = _apply_rdf_types(intrinsic, type_overrides)
    edges['type'] = None

    intrinsic = _format_date_fields(intrinsic)
    intrinsic = _remove_illegal_rdf_characters(intrinsic, illegal_characters, 'subject')
    intrinsic = _remove_illegal_rdf_characters(intrinsic, illegal_characters_intrinsic_object, 'object')
    edges = _remove_illegal_rdf_characters(edges, illegal_characters, 'subject')
    edges = _remove_illegal_rdf_characters(edges, illegal_characters, 'object')

    intrinsic = _remove_na_objects(intrinsic, drop_na_intrinsic_objects)
    edges = _remove_na_objects(edges, drop_na_edge_objects)

    _override_edge_name(edges, override_edge_name, key_seperator)

    intrinsic = intrinsic[['subject', 'predicate', 'object', 'type']]
    edges = edges[['subject', 'predicate', 'object', 'type']]

    return intrinsic, edges



from datetime import datetime
import unittest
import numpy as np

import pandas as pd
from pandas.testing import assert_frame_equal
from parameterized import parameterized

from dgraphpandas.strategies.vertical_helpers import (
    _compile_illegal_characters_regex, _expand_csv_edges, _ignore_fields,
    _join_key_fields, _add_dgraph_type_records, _break_up_intrinsic_and_edges,
    _apply_rdf_types, _format_date_fields, _override_edge_name,
    _remove_illegal_rdf_characters, _remove_na_objects, _rename_fields,
    _resolve_potential_callables)
from dgraphpandas.types import default_rdf_type


class VerticalHelpers(unittest.TestCase):

    @parameterized.expand([(None, ['csv_edge'])])
    def test_expand_csv_edges_null_parameters_exception(self, frame, csv_edges):
        '''
        Ensures when any of the parameters are invalid,
        an exception is raised
        '''
        with self.assertRaises(ValueError):
            _expand_csv_edges(frame, csv_edges)

    @parameterized.expand([
        (None,),
        ([],)
    ])
    def test_expand_csv_edges_no_csv_edges_same_frame_returned(self, csv_edges):
        '''
        Ensures when there are no csv_edges passed, then the same frame is returned.
        '''
        frame = pd.DataFrame(data={
            'subject': ['subject1', 'subject2', 'subject3'],
            'predicate': ['predicate1', 'predicate2', 'predicate3'],
            'object': ['object1', 'object2', 'object3']
        })

        result_frame = _expand_csv_edges(frame.copy(), csv_edges)
        assert_frame_equal(frame, result_frame)

    def test_expand_csv_edges_csv_edges_provided_but_no_eligible_values(self):
        '''
        Ensures when csv_edges are passed but the dataframe itself has no
        eligible values (which are in csv) then the same frame is returned
        '''
        csv_edges = ['predicate1']
        frame = pd.DataFrame(data={
            'subject': ['subject1', 'subject2', 'subject3'],
            'predicate': ['predicate1', 'predicate2', 'predicate3'],
            'object': ['object1', 'object2', 'object3']
        })

        result_frame = _expand_csv_edges(frame.copy(), csv_edges)

        expected = frame.sort_index().sort_values(by='subject')
        output = result_frame.sort_index().sort_values(by='subject')
        assert_frame_equal(expected, output)

    def test_expand_csv_edges_csv_edge_not_found_frame_unchanged(self):
        '''
        Ensures when the given csv edge does not actually exist
        on the frame, then the frame is unchanged
        '''
        csv_edges = ['predicate_no_exist']
        frame = pd.DataFrame(data={
            'subject': ['subject1', 'subject2', 'subject3'],
            'predicate': ['predicate1', 'predicate2', 'predicate3'],
            'object': ['object1', 'object2', 'object3']
        })

        result_frame = _expand_csv_edges(frame.copy(), csv_edges)

        expected = frame.sort_index().sort_values(by='subject')
        output = result_frame.sort_index().sort_values(by='subject')
        assert_frame_equal(expected, output)

    def test_expand_csv_edges_field_found(self):
        '''
        Ensures when the csv edge is found on the frame,
        the records are split out, new records are generated
        and the subject, predicate are copied out.

        Also ensures the default seperator is csv
        '''
        csv_edges = ['predicate2']
        frame = pd.DataFrame(data={
            'subject': ['subject1', 'subject2', 'subject3'],
            'predicate': ['predicate1', 'predicate2', 'predicate3'],
            'object': ['object1', 'object2a,object2b,object2c', 'object3']
        })

        expected = pd.DataFrame(data={
            'subject': ['subject1', 'subject3', 'subject2', 'subject2', 'subject2'],
            'predicate': ['predicate1', 'predicate3', 'predicate2', 'predicate2', 'predicate2'],
            'object': ['object1', 'object3', 'object2a', 'object2b', 'object2c']
        })

        result_frame = _expand_csv_edges(frame.copy(), csv_edges)

        actual = expected.reset_index(drop=True).sort_index()
        output = result_frame.reset_index(drop=True).sort_index()
        assert_frame_equal(actual, output)

    @parameterized.expand([('|',), ('#',)])
    def test_expand_csv_edges_alternative_seperator(self, seperator: str):
        '''
        Ensures when alternative separators are passed, they are broken
        up correctly.
        '''
        csv_edges = ['predicate2']
        frame = pd.DataFrame(data={
            'subject': ['subject1', 'subject2', 'subject3'],
            'predicate': ['predicate1', 'predicate2', 'predicate3'],
            'object': ['object1', seperator.join(['object2a', 'object2b', 'object2c']), 'object3']
        })

        expected = pd.DataFrame(data={
            'subject': ['subject1', 'subject3', 'subject2', 'subject2', 'subject2'],
            'predicate': ['predicate1', 'predicate3', 'predicate2', 'predicate2', 'predicate2'],
            'object': ['object1', 'object3', 'object2a', 'object2b', 'object2c']
        })

        result_frame = _expand_csv_edges(frame.copy(), csv_edges, seperator=seperator)

        actual = expected.reset_index(drop=True).sort_index()
        output = result_frame.reset_index(drop=True).sort_index()
        assert_frame_equal(actual, output)

    @parameterized.expand([
        (None, ['key'], 'key_seperator', 'type'),
        (pd.DataFrame(), None, 'key_seperator', 'type'),
        (pd.DataFrame(), ['key'], None, 'type'),
        (pd.DataFrame(), ['key'], 'key_seperator', None),
    ])
    def test_join_key_fields_null_parameters_exception(self, frame, key, key_seperator, type):
        '''
        Ensures when any of the input parameters are null, then
        an error is raised.
        '''
        with self.assertRaises(ValueError):
            _join_key_fields(frame, key, key_seperator, type)

    def test_join_key_fields_one_key_subject_created(self):
        '''
        Ensures when there is only one key, the
        subject is constructed and prefixed with type
        '''
        key = ['customer_id']
        key_seperator = '_'
        type = 'customer'
        frame = pd.DataFrame(data={
            'customer_id': [1, 2, 3],
            'predicate': ['age', 'age', 'age'],
            'object': [23, 45, 12],
        })

        expected = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_2', 'customer_3'],
            'predicate': ['age', 'age', 'age'],
            'object': [23, 45, 12],
        })

        result_frame = _join_key_fields(frame.copy(), key, key_seperator, type)

        actual = expected.reset_index(drop=True)[['subject', 'predicate', 'object']]
        output = result_frame.reset_index(drop=True)[['subject', 'predicate', 'object']]
        assert_frame_equal(actual, output)

    def test_join_key_fields_multiple_keys_subject_created(self):
        '''
        Ensures when there are multiple keys, then they are
        concatted together and the type is suffixed.
        '''
        key = ['customer_id', 'order_id']
        key_seperator = '_'
        type = 'order'
        frame = pd.DataFrame(data={
            'customer_id': [1, 2, 3],
            'order_id': [10, 22, 36],
            'predicate': ['value', 'value', 'value'],
            'object': [45.4, 76.9, 12.3],
        })

        expected = pd.DataFrame(data={
            'subject': ['order_1_10', 'order_2_22', 'order_3_36'],
            'predicate': ['value', 'value', 'value'],
            'object': [45.4, 76.9, 12.3],
        })

        result_frame = _join_key_fields(frame.copy(), key, key_seperator, type)

        actual = expected.reset_index(drop=True)[['subject', 'predicate', 'object']]
        output = result_frame.reset_index(drop=True)[['subject', 'predicate', 'object']]
        assert_frame_equal(actual, output)

    def test_add_dgraph_type_records_enabled_records_added(self):
        '''
        Ensures when add_dgraph_type_records is passed, then dgraph.type
        records are generated and appended to the DataFrame.
        '''
        type = 'customer'
        frame = pd.DataFrame(data={
            'subject': [1, 2, 3],
            'predicate': ['age', 'age', 'age'],
            'object': [23, 45, 12],
        })

        result_frame = _add_dgraph_type_records(frame.copy(), add_dgraph_type_records=True, type=type)

        expected = pd.DataFrame(data={
            'subject': [1, 2, 3, 1, 2, 3],
            'predicate': ['age']*3 + ['dgraph.type']*3,
            'object': [23, 45, 12] + [type]*3
        })

        actual = expected.reset_index(drop=True)[['subject', 'predicate', 'object']]
        output = result_frame.reset_index(drop=True)[['subject', 'predicate', 'object']]

        self.assertEqual(6, result_frame.shape[0])
        without_dgraph_type = result_frame[~result_frame['predicate'].isin(['dgraph.type'])]
        with_dgraph_type = result_frame[result_frame['predicate'].isin(['dgraph.type'])]
        self.assertEqual(3, without_dgraph_type.shape[0])
        self.assertEqual(3, with_dgraph_type.shape[0])
        self.assertEqual(with_dgraph_type['predicate'].values.tolist(), ['dgraph.type']*3)
        self.assertEqual(without_dgraph_type['predicate'].values.tolist(), ['age']*3)

        assert_frame_equal(actual, output)

    def test_add_dgraph_type_records_disabled_records_added(self):
        '''
        Ensures when add_dgraph_type_records is disabled, then
        the frame remains unchanged
        '''
        type = 'customer'
        frame = pd.DataFrame(data={
            'subject': [1, 2, 3],
            'predicate': ['age', 'age', 'age'],
            'object': [23, 45, 12],
        })

        result_frame = _add_dgraph_type_records(frame.copy(), add_dgraph_type_records=False, type=type)

        actual = frame.reset_index(drop=True)[['subject', 'predicate', 'object']]
        output = result_frame.reset_index(drop=True)[['subject', 'predicate', 'object']]
        assert_frame_equal(actual, output)

    def test_break_up_intrinsic_and_edges_null_parameter_exception(self):
        '''
        Ensures when parameters are null, then an exception is raised.
        '''
        with self.assertRaises(ValueError):
            _break_up_intrinsic_and_edges(None, ['edges'])

    @parameterized.expand([
        (None,),
        ([],)
    ])
    def test_break_up_intrinsic_and_edges_no_edges_only_intrinsic_and_default_edges(self, edges):
        '''
        Ensures when no edges have been defined, then only intrinsic has data
        and edges is the default.
        '''
        frame = pd.DataFrame(data={
            'subject': [1, 2, 3],
            'predicate': ['age', 'age', 'age'],
            'object': [23, 45, 12],
        })

        intrinsic, edges = _break_up_intrinsic_and_edges(frame.copy(), edges)

        actual = frame.reset_index(drop=True)[['subject', 'predicate', 'object']]
        output = intrinsic.reset_index(drop=True)[['subject', 'predicate', 'object']]
        assert_frame_equal(actual, output)

        self.assertTrue(edges.empty)
        self.assertEqual(edges.columns.tolist(), ['subject', 'predicate', 'object', 'type'])

    def test_break_up_intrinsic_and_edges_edges_exist_default_id_stripped(self):
        '''
        Ensures when edges exist they are split into edges frame
        and by default _id is stripped
        '''

        frame = pd.DataFrame(data={
            'subject': ['1', '2', '3', '34'],
            'predicate': ['age', 'location_id', 'age', 'class_id'],
            'object': ['23', 'london', '12', 'first'],
        })

        expected_intrinsic = pd.DataFrame(data={
            'subject': ['1', '3'],
            'predicate': ['age', 'age'],
            'object': ['23', '12'],
        })

        expected_edges = pd.DataFrame(data={
            'subject': ['2', '34'],
            'predicate': ['location', 'class'],
            'object': ['london', 'first'],
        })

        intrinsic, edges = _break_up_intrinsic_and_edges(frame.copy(), ['location_id', 'class_id'])

        expected_intrinsic = expected_intrinsic.reset_index(drop=True)[['subject', 'predicate', 'object']]
        actual_intrinsic = intrinsic.reset_index(drop=True)[['subject', 'predicate', 'object']]
        assert_frame_equal(expected_intrinsic, actual_intrinsic)

        expected_edges = expected_edges.reset_index(drop=True)[['subject', 'predicate', 'object']]
        actual_intrinsic = edges.reset_index(drop=True)[['subject', 'predicate', 'object']]
        assert_frame_equal(expected_edges, actual_intrinsic)

    @parameterized.expand([
        (True,),
        (False,)
    ])
    def test_break_up_intrinsic_and_edges_edges_exist_id_stripped_passed(self, strip_id_from_edge_names: bool):
        '''
        Ensures when the strip_id_from_edge_names parameter is passed
        the edge names are stripped accordingly.
        '''
        frame = pd.DataFrame(data={
            'subject': ['1', '2', '3', '34'],
            'predicate': ['age', 'location_id', 'age', 'class_id'],
            'object': ['23', 'london', '12', 'first'],
        })

        expected_intrinsic = pd.DataFrame(data={
            'subject': ['1', '3'],
            'predicate': ['age', 'age'],
            'object': ['23', '12'],
        })

        expected_edges = pd.DataFrame(data={
            'subject': ['2', '34'],
            'predicate': (['location', 'class'] if strip_id_from_edge_names else ['location_id', 'class_id']),
            'object': ['london', 'first'],
        })

        intrinsic, edges = _break_up_intrinsic_and_edges(frame.copy(), ['location_id', 'class_id'], strip_id_from_edge_names)

        expected_intrinsic = expected_intrinsic.reset_index(drop=True)[['subject', 'predicate', 'object']]
        actual_intrinsic = intrinsic.reset_index(drop=True)[['subject', 'predicate', 'object']]
        assert_frame_equal(expected_intrinsic, actual_intrinsic)

        expected_edges = expected_edges.reset_index(drop=True)[['subject', 'predicate', 'object']]
        actual_edges = edges.reset_index(drop=True)[['subject', 'predicate', 'object']]
        assert_frame_equal(expected_edges, actual_edges)

    @parameterized.expand([
        (None, {'column1': '<xs:int>'}),
        (pd.DataFrame(), None)
    ])
    def test_apply_rdf_types_nullparameters_exception(self, frame, types):
        '''
        Ensures when either of the parameters is null, then a value exception is raised
        '''
        with self.assertRaises(ValueError):
            _apply_rdf_types(frame, types)

    def test_apply_rdf_types_some_columns_to_map(self):
        '''
        Ensures when some of the columns have matching RDF types
        they are mapped and the unmapped are given the default rdf type
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'dob', 'weight'],
            'object':  ['black', '1921-05-03T00:00:00', '50']
        })

        types = {
            'dob': 'datetime64',
            'weight': 'int32'
        }

        expected_frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'dob', 'weight'],
            'object':  ['black', '1921-05-03T00:00:00', '50'],
            'type':  [default_rdf_type, '<xs:dateTime>', '<xs:int>']
        })

        result_frame = _apply_rdf_types(frame.copy(), types)
        assert_frame_equal(result_frame, expected_frame)

    def test_apply_rdf_types_all_columns_to_map(self):
        '''
        Ensures when all columns have a mapping, they are mapped to their
        corresponding rdf type
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'dob', 'weight'],
            'object':  ['black', '1921-05-03T00:00:00', '50']
        })

        types = {
            'hair_colour': 'object',
            'dob': 'datetime64',
            'weight': 'int32'
        }

        expected_frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'dob', 'weight'],
            'object':  ['black', '1921-05-03T00:00:00', '50'],
            'type':  ['<xs:string>', '<xs:dateTime>', '<xs:int>']
        })

        result_frame = _apply_rdf_types(frame.copy(), types)
        assert_frame_equal(result_frame, expected_frame)

    def test_apply_rdf_types_no_columns_to_map(self):
        '''
        Ensures when there are no mappings defined, then all columns
        are given the default rdf type
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'dob', 'weight'],
            'object':  ['black', '1921-05-03T00:00:00', '50']
        })

        types = {}

        expected_frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'dob', 'weight'],
            'object':  ['black', '1921-05-03T00:00:00', '50'],
            'type':  [default_rdf_type]*3
        })

        result_frame = _apply_rdf_types(frame.copy(), types)
        assert_frame_equal(result_frame, expected_frame)

    def test_format_date_fields_null_frame_error(self):
        '''
        Ensures when parameters are null, then an error is raised
        '''
        with self.assertRaises(ValueError):
            _format_date_fields(None)

    def test_format_date_fields_no_date_fields(self):
        '''
        Ensures when there are no date fields,
        then the frame is unchanged
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'height', 'weight'],
            'object':  ['black', '172', '50'],
            'type':  ['<xs:string>', '<xs:int>', '<xs:int>']
        })

        result_frame = _format_date_fields(frame.copy())
        assert_frame_equal(frame, result_frame)

    def test_format_date_fields_date_fields_exist(self):
        '''
        Ensures when there are fields marked with datetime type
        then they are converted into ISO format.
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'dob', 'weight'],
            'object':  ['black', datetime(2021, 1, 4), '50'],
            'type':  ['<xs:string>', '<xs:dateTime>', '<xs:int>']
        })

        expected = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'dob', 'weight'],
            'object':  ['black', '2021-01-04T00:00:00', '50'],
            'type':  ['<xs:string>', '<xs:dateTime>', '<xs:int>']
        })

        result_frame = _format_date_fields(frame.copy())

        result_frame = result_frame.sort_values(by='predicate').reset_index(drop=True)
        expected = expected.sort_values(by='predicate').reset_index(drop=True)
        assert_frame_equal(result_frame, expected)

    def test_format_date_fields_date_fields_exist_but_not_datetime(self):
        '''
        Ensures when a date field is provided but it's not
        an actual datetime object, an error is raised
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'dob', 'weight'],
            'object':  ['black', '2021 Jan 21', '50'],
            'type':  ['<xs:string>', '<xs:dateTime>', '<xs:int>']
        })

        with self.assertRaises(AttributeError):
            _format_date_fields(frame)

    def test_compile_illegal_characters_regex_nonecharacters(self):
        '''
        Ensure when none characters are passed, then none
        is returned
        '''
        self.assertIsNone(_compile_illegal_characters_regex(None))

    @parameterized.expand([
        (['@'], '@'),
        (['$', '@'], '$|@'),
        (['\\^', '\\#', '\\~', '\\+'], '\\^|\\#|\\~|\\+'),
    ])
    def test_compile_illegal_characters_regex_characterspassed(self, characters, expected_regex):
        '''
        Ensures when the given regex is provided
        then the predefined regex pattern is constructed.
        '''
        regex = _compile_illegal_characters_regex(characters)
        self.assertEqual(expected_regex, regex.pattern)

    @parameterized.expand([
        (None, 'subject'),
        (pd.DataFrame(), None),
        (pd.DataFrame(), ''),
    ])
    def test_remove_illegal_rdf_characters_null_parameters(self, frame, field):
        '''
        Ensures when parameters are null, then an error is raised
        '''
        with self.assertRaises(ValueError):
            _remove_illegal_rdf_characters(frame, ['$', '@'], field)

    def test_remove_illegal_rdf_characters_illegalcharacters_found(self):
        '''
        Ensures when there are illegal characters found
        in the field, then they are removed
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_2'],
            'predicate': ['description', 'description'],
            'object': ['hello @world#', 'this is a fine string']
        })
        illegal_characters = ['@', '#']
        field = 'object'

        expected_frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_2'],
            'predicate': ['description', 'description'],
            'object': ['hello world', 'this is a fine string']
        })

        result_frame = _remove_illegal_rdf_characters(frame.copy(), illegal_characters, field)
        assert_frame_equal(expected_frame, result_frame)

    def test_remove_illegal_rdf_characters_illegalcharacters_notfound(self):
        '''
        Ensures when no illegal characters are found, then
        the frame is unchanged.
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_2'],
            'predicate': ['description', 'description'],
            'object': ['hello world', 'this is a fine string']
        })
        illegal_characters = ['@', '#']
        field = 'object'

        result_frame = _remove_illegal_rdf_characters(frame.copy(), illegal_characters, field)
        assert_frame_equal(frame, result_frame)

    def test_remove_illegal_rdf_characters_only_one_field(self):
        '''
        Ensures when there are illegal characters in both subject and object
        but we are only working on one of them right now then the target
        field is cleaned but the other is untouched.
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1@@', 'custo#mer_2'],
            'predicate': ['description', 'description'],
            'object': ['hello @world#', 'this is a fine string']
        })
        illegal_characters = ['@', '#']
        field = 'subject'

        expected_frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_2'],
            'predicate': ['description', 'description'],
            'object': ['hello @world#', 'this is a fine string']
        })

        result_frame = _remove_illegal_rdf_characters(frame.copy(), illegal_characters, field)
        assert_frame_equal(expected_frame, result_frame)

    def test_remove_na_objects_null_parameters(self):
        '''
        Ensures when parameters are null, then an error is raised
        '''
        with self.assertRaises(ValueError):
            _remove_na_objects(None)

    @parameterized.expand([
        (np.nan,),
        (None,),
        (pd.NA,),
        ('NA',)
    ])
    def test_remove_na_objects_false(self, na_value):
        '''
        Ensures when there are NA values but the drop_na flag
        is set to false, then the frame is unchanged
        '''
        frame = pd.DataFrame(data={
            'subject': ['subject1', 'subject2', 'subject3'],
            'predicate': ['predicate1', 'predicate2', 'predicate3'],
            'object': ['object1', na_value, 'object3']
        })

        result_frame = _remove_na_objects(frame.copy(), drop_na=False)
        assert_frame_equal(frame, result_frame)

    @parameterized.expand([
        (np.nan,),
        (None,),
        (pd.NA,)
    ])
    def test_remove_na_objects_true(self, na_value):
        '''
        Ensures when drop_na is set to true then
        those records are removed from the frame.
        '''
        frame = pd.DataFrame(data={
            'subject': ['subject1', 'subject2', 'subject3'],
            'predicate': ['predicate1', 'predicate2', 'predicate3'],
            'object': ['object1', na_value, 'object3']
        })

        expected_frame = pd.DataFrame(data={
            'subject': ['subject1', 'subject3'],
            'predicate': ['predicate1', 'predicate3'],
            'object': ['object1', 'object3']
        })

        result_frame = _remove_na_objects(frame.copy(), drop_na=True)

        result_frame = result_frame.reset_index(drop=True).sort_index()
        expected_frame = expected_frame.reset_index(drop=True).sort_index()
        assert_frame_equal(expected_frame, result_frame)

    @parameterized.expand([
        (None, 'sep'),
        (pd.DataFrame(), None),
    ])
    def test_override_edge_name_null_parameters(self, edges, key_seperator):
        '''
        Ensures when parameters is none, then errors are raised
        '''
        with self.assertRaises(ValueError):
            _override_edge_name(edges, {}, key_seperator)

    @parameterized.expand([
        (None,),
        ({},),
    ])
    def test_override_edge_name_no_overrides_default_apply(self, override_edge_name):
        '''
        Ensures when no override_edge_name has been provided
        then the default add predicate to object is applied.
        '''
        edges = pd.DataFrame(data={
            'customer_id': ['customer_1', 'customer_2', 'customer_3'],
            'predicate': ['location', 'location', 'location'],
            'object': [23, 45, 12],
        })
        key_seperator = '_'

        expected_edges = pd.DataFrame(data={
            'customer_id': ['customer_1', 'customer_2', 'customer_3'],
            'predicate': ['location', 'location', 'location'],
            'object': ['location_23', 'location_45', 'location_12'],
        })

        edges = _override_edge_name(edges.copy(), override_edge_name, key_seperator)
        assert_frame_equal(expected_edges, edges)

    def test_override_edge_name_override_provided(self):
        '''
        Ensures when both a target_node_type and predicate and provided
        within an override they are applied to the edge and any not defined
        in override and left untouched
        '''
        edges = pd.DataFrame(data={
            'customer_id': ['customer_1', 'customer_2', 'customer_3', 'customer_1'],
            'predicate': ['location', 'location', 'location', 'purchased'],
            'object': [23, 45, 12, 'apple'],
        })
        key_seperator = '_'
        override_edge_name = {
            'location': {
                'target_node_type': 'habitat',
                'predicate': 'secondary_location'
            }
        }

        expected_edges = pd.DataFrame(data={
            'customer_id': ['customer_1', 'customer_2', 'customer_3', 'customer_1'],
            'predicate': ['secondary_location', 'secondary_location', 'secondary_location', 'purchased'],
            'object': ['habitat_23', 'habitat_45', 'habitat_12', 'purchased_apple'],
        })

        edges = _override_edge_name(edges.copy(), override_edge_name, key_seperator)
        assert_frame_equal(expected_edges, edges)

    def test_override_edge_name_override_provided_no_predicate(self):
        '''
        Ensures when a target_node_type is provided but no override predicate is provided
        then the predicate is not touched and only the target_node_type
        is changed.
        '''
        edges = pd.DataFrame(data={
            'customer_id': ['customer_1', 'customer_2', 'customer_3', 'customer_1'],
            'predicate': ['location', 'location', 'location', 'purchased'],
            'object': [23, 45, 12, 'apple'],
        })
        key_seperator = '_'
        override_edge_name = {
            'location': {
                'target_node_type': 'habitat'
            }
        }

        expected_edges = pd.DataFrame(data={
            'customer_id': ['customer_1', 'customer_2', 'customer_3', 'customer_1'],
            'predicate': ['location', 'location', 'location', 'purchased'],
            'object': ['habitat_23', 'habitat_45', 'habitat_12', 'purchased_apple'],
        })

        edges = _override_edge_name(edges.copy(), override_edge_name, key_seperator)
        assert_frame_equal(expected_edges, edges)

    def test_ignore_fields_null_frame(self):
        '''
        Ensures when the frame is null, then an error is raised.
        '''
        with self.assertRaises(ValueError):
            _ignore_fields(None, ['internal_customer_id'])

    @parameterized.expand([
        (None,),
        ([],),
    ])
    def test_ignore_no_ignorefields(self, ignore_fields):
        '''
        Ensures whn no ignore fields are provided, then the
        frame is untouched.
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'height', 'weight'],
            'object':  ['black', '172', '50'],
            'type':  ['<xs:string>', '<xs:int>', '<xs:int>']
        })

        result = _ignore_fields(frame.copy(), ignore_fields)
        assert_frame_equal(result, frame)

    def test_ignore_fields_provided_but_not_in_frame(self):
        '''
        Ensures when ignore fields are provided but
        they do not exist in the dataframe, then
        the frame is untouched.
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'height', 'weight'],
            'object':  ['black', '172', '50'],
            'type':  ['<xs:string>', '<xs:int>', '<xs:int>']
        })
        ignore_fields = ['dob']

        result = _ignore_fields(frame.copy(), ignore_fields)
        assert_frame_equal(result, frame)

    def test_ignore_fields_provided_and_in_frame(self):
        '''
        Ensures when ignore fields are provided and
        the predicate does exist then it is removed.
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'height', 'weight'],
            'object':  ['black', '172', '50'],
            'type':  ['<xs:string>', '<xs:int>', '<xs:int>']
        })
        ignore_fields = ['height']

        expected_frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'weight'],
            'object':  ['black', '50'],
            'type':  ['<xs:string>', '<xs:int>']
        })

        result = _ignore_fields(frame.copy(), ignore_fields)

        expected_frame = expected_frame.reset_index(drop=True).sort_index()
        result = result.reset_index(drop=True).sort_index()
        assert_frame_equal(result, expected_frame)

    def test_resolve_potential_callables_null_frame(self):
        '''
        Ensures when the frame is null, then an error is raised
        '''
        potential_callables = {'key': ['key_id']}
        with self.assertRaises(ValueError):
            _resolve_potential_callables(None, potential_callables)

    def test_resolve_potential_callables_no_callables(self):
        '''
        Ensures when there are no callables passed, the dictionary
        is unchanged
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['cid', 'location_id', 'weight'],
            'object':  [1420, '45', '50']
        })

        potential_callables = {
            'key': ['cid'],
            'edges': ['location_id']
        }

        result_callables = _resolve_potential_callables(frame, potential_callables)
        self.assertEqual(potential_callables, result_callables)

    def test_resolve_potential_callables_callablesprovided(self):
        '''
        Ensures when a callable is provided it's applied against the given frame
        and returned
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['cid', 'location_id', 'weight'],
            'object':  [1420, '45', '50']
        })

        potential_callables = {
            'key': ['cid'],
            'edges': lambda frame: frame.loc[frame['predicate'].str.endswith('_id'), 'predicate'].unique().tolist()
        }

        expected_result = {
            'key': ['cid'],
            'edges': ['location_id']
        }

        result_callables = _resolve_potential_callables(frame, potential_callables)
        self.assertEqual(expected_result, result_callables)

    def test_rename_fields_null_frame(self):
        '''
        Ensures when parameters are null, then an error is raised
        '''
        pre_rename = {'loc_id': 'location_id'}
        with self.assertRaises(ValueError):
            _rename_fields(None, pre_rename)

    @parameterized.expand([
        (None,),
        ({},),
    ])
    def test_rename_fields_no_prerename(self, rename_fields):
        '''
        Ensures when no rename fields are provided,
        then the frame is untouched.
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'height', 'weight'],
            'object':  ['black', '172', '50'],
            'type':  ['<xs:string>', '<xs:int>', '<xs:int>']
        })

        result = _rename_fields(frame.copy(), rename_fields)
        assert_frame_equal(result, frame)

    def test_rename_fields_provided(self):
        '''
        Ensures when rename fields are provided,
        they are applied to the dataframe.
        '''
        frame = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair_colour', 'height', 'mass'],
            'object':  ['black', '172', '50'],
            'type':  ['<xs:string>', '<xs:int>', '<xs:int>']
        })
        rename_fields = {
            'hair_colour': 'hair',
            'mass': 'weight'
        }

        expected = pd.DataFrame(data={
            'subject': ['customer_1', 'customer_1', 'customer_1'],
            'predicate':  ['hair', 'height', 'weight'],
            'object':  ['black', '172', '50'],
            'type':  ['<xs:string>', '<xs:int>', '<xs:int>']
        })

        result = _rename_fields(frame.copy(), rename_fields)
        assert_frame_equal(result, expected)



import logging
from typing import List, Tuple

import pandas as pd
from dgraphpandas.types import default_rdf_type

logger = logging.getLogger(__name__)


def _generate_intrinsic(intrinsic: pd.DataFrame) -> List[str]:
    intrinsic['subject'] = intrinsic['subject'].astype(str)
    intrinsic['predicate'] = intrinsic['predicate'].astype(str)
    intrinsic['object'] = intrinsic['object'].astype(str)

    intrinsic['type'] = intrinsic['type'].fillna(default_rdf_type)
    intrinsic['dql'] = '<' + intrinsic['subject'] + '>' + ' ' + '<' + intrinsic['predicate'] + \
        '>' + ' ' + '"' + intrinsic['object'] + '"' + '^^' + intrinsic['type'] + ' .'
    intrinsic.drop(columns=['subject', 'predicate', 'object', 'type'], inplace=True)
    intrinsic = intrinsic['dql'].values.tolist()
    return intrinsic


def _generate_edges(edges: pd.DataFrame) -> List[str]:
    edges['subject'] = edges['subject'].astype(str)
    edges['predicate'] = edges['predicate'].astype(str)
    edges['object'] = edges['object'].astype(str)

    edges['dql'] = '<' + edges['subject'] + '>' + ' ' + '<' + edges['predicate'] + '>' + ' ' + '<' + edges['object'] + '>' + ' .'
    edges.drop(columns=['subject', 'predicate', 'object', 'type'], inplace=True)
    edges = edges['dql'].values.tolist()
    return edges


def generate_upserts(
        intrinsic: pd.DataFrame,
        edges: pd.DataFrame, drop_na_objects=True) -> Tuple[List[str], List[str]]:

    if intrinsic is None:
        raise ValueError('intrinsic')
    if edges is None:
        raise ValueError('edges')

    required_intrinsic_fields = ['subject', 'predicate', 'object', 'type']
    for col in required_intrinsic_fields:
        if col not in intrinsic.columns:
            raise ValueError(f'{col} is not within intrinsic columns {intrinsic.columns}')

    required_edges_fields = ['subject', 'predicate', 'object']
    for col in required_edges_fields:
        if col not in edges.columns:
            raise ValueError(f'{col} is not within edges columns {edges.columns}')

    if drop_na_objects:
        logger.debug('Dropping NA Objects from intrinsic')
        intrinsic.dropna(subset=['object'], inplace=True)

    intrinsic_upserts = _generate_intrinsic(intrinsic)
    edge_upserts = _generate_edges(edges)

    return (intrinsic_upserts, edge_upserts)



import os

from typing import Dict
import logging
import numpy as np

logger = logging.getLogger(__name__)

default_rdf_type = os.environ.get('DEFAULT_RDF_TYPE', '<xs:string>')


numpy_str_to_rdf_types = {
    'object': '<xs:string>',
    'O': '<xs:string>',
    'int32': '<xs:int>',
    'int64': '<xs:int>',
    'float32': '<xs:float>',
    'float64': '<xs:float>',
    'datetime64': '<xs:dateTime>',
    '<M8[ns]': '<xs:dateTime>',
    'bool': '<xs:bool>'
}

numpy_to_rdf_types = {
    key: np.dtype(key) for key, _ in numpy_str_to_rdf_types.items()
}


def find_rdf_types(types: Dict[str, str]):

    if types is None:
        raise ValueError('types')

    resolved_types = {}
    for col, type in types.items():
        resolved_types[col] = numpy_str_to_rdf_types.get(type, default_rdf_type)
    return resolved_types



import unittest
from parameterized import parameterized

import pandas as pd

from dgraphpandas.writers.upserts import generate_upserts


class UpsertTests(unittest.TestCase):

    @parameterized.expand([
        (None, pd.DataFrame()),
        (pd.DataFrame(), None),
    ])
    def test_generate_upserts_null_parameters(self, intrinsic, edges):
        '''
        Ensures when null parameters are provided, then
        an error is raised
        '''
        with self.assertRaises(ValueError):
            generate_upserts(intrinsic, edges)

    @parameterized.expand([
        ###
        (
            'subject_no_exist',
            pd.DataFrame(data={
                        'predicate': ['age', 'hair'],
                        'object': [23, 'black'],
                        'type': ['<xs:int>', '<xs:string>']
            }),
        ),
        ###
        (
            'predicate_no_exist',
            pd.DataFrame(data={
                        'subject': ['customer_1', 'customer_2'],
                        'object': [23, 'black'],
                        'type': ['<xs:int>', '<xs:string>']
            }),
        ),
        ###
        (
            'object_no_exist',
            pd.DataFrame(data={
                        'subject': ['customer_1', 'customer_2'],
                        'predicate': ['age', 'hair'],
                        'type': ['<xs:int>', '<xs:string>']
            }),
        ),
        ###
        (
            'type_no_exist',
            pd.DataFrame(data={
                        'subject': ['customer_1', 'customer_2'],
                        'predicate': ['age', 'hair'],
                        'object': [23, 'black'],
            }),
        )
    ])
    def test_generate_upsert_intrinsic_has_bad_columns(self, name, intrinsic):
        '''
        Ensures when the intrinsic frame does not have the correct
        columns, then an exception is raised
        '''
        edges = pd.DataFrame(columns=['subject', 'predicate', 'object', 'type'])
        with self.assertRaises(ValueError):
            generate_upserts(intrinsic, edges)

    @parameterized.expand([
        ###
        (
            'subject_no_exist',
            pd.DataFrame(data={
                        'predicate': ['location', 'knows'],
                        'object': ['loc34', 'customer_2'],
            }),
        ),
        ###
        (
            'predicate_no_exist',
            pd.DataFrame(data={
                        'subject': ['customer_1', 'customer_2'],
                        'object': ['loc34', 'customer_2'],
            }),
        ),
        ###
        (
            'object_no_exist',
            pd.DataFrame(data={
                        'subject': ['customer_1', 'customer_2'],
                        'predicate': ['location', 'knows'],
            }),
        )
    ])
    def test_generate_upsert_edges_has_bad_columns(self, name, edges):
        '''
        Ensures when the edges frame does not have the correct
        columns, then an exception is raised
        '''
        intrinsic = pd.DataFrame(columns=['subject', 'predicate', 'object', 'type'])
        with self.assertRaises(ValueError):
            generate_upserts(intrinsic, edges)

    @parameterized.expand([
        ###
        (
            'single_record',
            pd.DataFrame(data={
                        'subject': ['customer_1'],
                        'predicate': ['age'],
                        'object': [23],
                        'type': ['<xs:int>']
            }),
            ['<customer_1> <age> "23"^^<xs:int> .']
        ),
        ###
        (
            'multiple_records',
            pd.DataFrame(data={
                        'subject': ['customer_1', 'customer_2', 'customer_1', 'customer_3'],
                        'predicate': ['age', 'hair', 'dob', 'weight'],
                        'object': [23, 'black', '20210302T00:00:00', 1.32],
                        'type': ['<xs:int>', '<xs:string>', '<xs:dateTime>', '<xs:float>']
            }),
            [
                '<customer_1> <age> "23"^^<xs:int> .',
                '<customer_2> <hair> "black"^^<xs:string> .',
                '<customer_1> <dob> "20210302T00:00:00"^^<xs:dateTime> .',
                '<customer_3> <weight> "1.32"^^<xs:float> .'
            ]
        )
    ])
    def test_generate_upsert_only_intrinsic(self, name, intrinsic, expected_output):
        '''
        Ensures when only intrinsic data is passed, we generate upserts only for intrinsic.
        '''
        edges = pd.DataFrame(columns=['subject', 'predicate', 'object', 'type'])

        dql_intrinsic, dql_edges = generate_upserts(intrinsic, edges)

        self.assertEqual(len(dql_edges), 0)
        self.assertEqual(dql_intrinsic, expected_output)

    @parameterized.expand([
        ###
        (
            'single_record',
            pd.DataFrame(data={
                        'subject': ['customer_1'],
                        'predicate': ['location'],
                        'object': ['loc_32'],
                        'type': [None]
            }),
            ['<customer_1> <location> <loc_32> .']
        ),
        ###
        (
            'multiple_records',
            pd.DataFrame(data={
                        'subject': ['customer_1', 'customer_3', 'customer_2'],
                        'predicate': ['location', 'group', 'location'],
                        'object': ['loc_32', 'group_89', 'loc_90'],
                        'type': [None]*3
            }),
            [
                '<customer_1> <location> <loc_32> .',
                '<customer_3> <group> <group_89> .',
                '<customer_2> <location> <loc_90> .'
            ]
        )
    ])
    def test_generate_upsert_only_edges(self, name, edges, expected_output):
        '''
        Ensures when there are only edges, the expected output is generated
        '''
        intrinsic = pd.DataFrame(columns=['subject', 'predicate', 'object', 'type'])

        dql_intrinsic, dql_edges = generate_upserts(intrinsic, edges)

        self.assertEqual(len(dql_intrinsic), 0)
        self.assertEqual(dql_edges, expected_output)

    @parameterized.expand([
        ###
        (
            'single_intrinsic_and_edge',
            pd.DataFrame(data={
                        'subject': ['customer_1'],
                        'predicate': ['age'],
                        'object': [23],
                        'type': ['<xs:int>']
            }),
            pd.DataFrame(data={
                        'subject': ['customer_1'],
                        'predicate': ['location'],
                        'object': ['loc_32'],
                        'type': [None]
            }),
            [
                '<customer_1> <age> "23"^^<xs:int> .',
                '<customer_1> <location> <loc_32> .'
            ]
        ),
        ###
        (
            'multiple_records',
            pd.DataFrame(data={
                        'subject': ['customer_1', 'customer_2', 'customer_1', 'customer_3'],
                        'predicate': ['age', 'hair', 'dob', 'weight'],
                        'object': [23, 'black', '20210302T00:00:00', 1.32],
                        'type': ['<xs:int>', '<xs:string>', '<xs:dateTime>', '<xs:float>']
            }),
            pd.DataFrame(data={
                        'subject': ['customer_1', 'customer_3', 'customer_2'],
                        'predicate': ['location', 'group', 'location'],
                        'object': ['loc_32', 'group_89', 'loc_90'],
                        'type': [None]*3
            }),
            [
                '<customer_1> <age> "23"^^<xs:int> .',
                '<customer_2> <hair> "black"^^<xs:string> .',
                '<customer_1> <dob> "20210302T00:00:00"^^<xs:dateTime> .',
                '<customer_3> <weight> "1.32"^^<xs:float> .',
                '<customer_1> <location> <loc_32> .',
                '<customer_3> <group> <group_89> .',
                '<customer_2> <location> <loc_90> .'
            ]
        )
    ])
    def test_generate_upsert_both_intrinsic_edges(self, name, intrinsic, edges, expected_output):
        '''
        Ensures when there are both intrinsic and edges, the expected output is generated
        '''
        dql_intrinsic, dql_edges = generate_upserts(intrinsic, edges)
        all_dql = dql_intrinsic + dql_edges
        self.assertEqual(all_dql, expected_output)




from typing import Dict, Any


def get_from_config(key: str, config: Dict[str, Any], default: Any = None, **kwargs) -> Any:
    '''
    For the given configuration key,
    first try to pull from kwargs (if there is an override)
    and secondly pull from the passed config dictionary
    else return default
    '''
    if not key:
        raise ValueError(key)
    if not config:
        raise ValueError(config)

    return kwargs.get(key, config.get(key, default))



import logging
from typing import Any, Dict, List, Callable, Union

import pandas as pd

from dgraphpandas.config import get_from_config
from dgraphpandas.strategies.vertical import vertical_transform

logger = logging.getLogger(__name__)


def horizontal_transform(
        frame: Union[str, pd.DataFrame],
        config: Dict[str, Any],
        config_file_key: str,
        **kwargs):
    '''
    Horizontally Transform a Pandas DataFrame into Intrinsic and Edge DataFrames.
    '''

    file_config: Dict[str, Any] = config['files'][config_file_key]
    type_overrides: Dict[str, str] = get_from_config('type_overrides', file_config, {}, **(kwargs))
    subject_fields: Union[List[str], Callable[..., List[str]]] = get_from_config('subject_fields', file_config, **(kwargs))

    if isinstance(frame, str):
        logger.debug(f'Reading file {frame}')
        frame = pd.read_csv(frame)

    '''
    Ensure that object values have the correct type according to type_overrides.
    For example, when pandas reads a csv and detects a numerical value it may decide to
    represent them as a float e.g 10.0 so when it's melted into a string it will show as such
    But we really want the value to be just 10 so it matches the corresponding rdf type.
    Therefore before we melt the frame, we enforce these columns have the correct form.
    '''
    logger.debug('Applying Type Overrides %s', type_overrides)
    for col, type in type_overrides.items():
        try:
            logger.debug(f'Converting {col} to {type}')
            frame[col] = frame[col].astype(type)
        except ValueError:
            logger.exception(
                f'''
                Could not convert {col} to {type}.
                Please confirm that the values in the {col} series are convertable to {type}.
                A common scenario here is when we have NA values but the target type does not support them.
                ''')
            exit()

    '''
    Pivot the Horizontal DataFrame based on the given key (subject).
    Change the frame to be 3 columns with triples: subject, predicate, object

    This changes the horizontal frame into a vertical frame as this more closely
    resembles rdf triples.
    '''
    logger.debug(f'Melting frame with subject: {subject_fields}')
    frame = frame.melt(
        id_vars=subject_fields,
        var_name='predicate',
        value_name='object')

    return vertical_transform(frame, config, config_file_key, **(kwargs))



from setuptools import setup, find_packages
from dgraphpandas import __name__, __version__

with open("README.md", "r") as desc:
    long_description = desc.read()

    with open('requirements.txt', 'r') as f:
        requirements = f.read()
        requirements = requirements.split('\n')

        setup(
            name=__name__,
            version=__version__,
            packages=find_packages(),
            install_requires=requirements,
            long_description=long_description,
            long_description_content_type="text/markdown",
            description="Transform Pandas DataFrames into Exports to be sent to DGraph",
            url="https://github.com/kiran94/dgraphpandas",
            entry_points={
                'console_scripts': [
                    'dgraphpandas = dgraphpandas.__main__:main'
                ]
            },
            keywords=['dgraph', 'pandas', 'rdf'],
            python_requires='>=3.6',
            license='MIT'
        )
